angular.module('sdbgui', ['ngRoute', 'ngCookies', 'ngAnimate', 'angularSlashDB'])
.config(['$routeProvider', '$locationProvider',
        function config($routeProvider, $locationProvider) {
            $routeProvider
                .when('/', {
                    templateUrl: '/login-static/app/welcome.template.html',
                    controller: function(){
                        window.location = '/';
                    }
                })
                .when('/login', {
                    templateUrl: '/login-static/app/welcome.template.html',
                    controller: 'LoginController',
                    controllerAs: 'ctrl'
                })
                .when('/login/sso/:idp_id', {
                    templateUrl: '/login-static/app/in-progress.template.html',
                    controller: 'SSOJwtController',
                    controllerAs: 'ctrl'
                })
                .when('/login/callback', {
                    templateUrl: '/login-static/app/in-progress.template.html',
                    controller: 'JwtCallbackController'
                })
                .when('/login/callback/failed', {
                    templateUrl: '/login-static/app/failed.template.html',
                    controller: 'SSOFailedController'
                })

            $locationProvider.html5Mode({
                enabled: true,
                requireBase: false
            });
        }
    ]
)

.factory('utils', ['$window', '$location', function($window, $location) {
    function base64URL (string) {
            return string.toString(CryptoJS.enc.Base64)
                .replace(/=/g, '')
                .replace(/\+/g, '-')
                .replace(/\//g, '_');
        }

    function goToReferrer (ref) {

        let stored_ref = $window.sessionStorage.getItem('ssoApp.login_referrer');
        // redirect back to /login/* may create loops
        // also redirect back to /login is not useful
        if (stored_ref) {
            let url = new URL(stored_ref)
            if (url.pathname.startsWith('/login')) {
                stored_ref = null
            }
        }

        // priority: arg, storage, '/'
        ref = (ref) ? ref : stored_ref
        ref = (ref) ? ref : '/';

        window.location = ref;
    }

    return {
        generateRandomString: function (length) {
            let text = "";
            let possible = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-._~";
            for (let i = 0; i < length; i++) {
                text += possible.charAt(Math.floor(Math.random() * possible.length));
            }
            return text;
        },
        generateCodeChallenge: function (code_verifier) {
            return base64URL(CryptoJS.SHA256(code_verifier));
        },

        base64URL: function (string) {
            return base64URL(string);
        },
        loginSuccessful: function() {
            // mark session exists
            // used only to show popup, no security related code based on it
            $window.localStorage.setItem('sdb.loggedIn', 'true');
            $window.localStorage.removeItem('sdb.sessionLost');

            const sendChannel = new BroadcastChannel('app-data');
            sendChannel.postMessage('sdb.loggedIn');

            if ('close' in $location.search()) {
                $window.close()
            }
            goToReferrer()
        },
        loginFailed: function() {
            goToReferrer('/login')
        },
        hashToObject: function (hash) {
            // Predefine regexs for parsing hash
            let plus2space = /\+/g;
            let paramSplit = /([^&=]+)=?([^&]*)/g;

            // Remove the leading hash
            // let fragment = hash.substring(1);

            let obj = {};

            // Loop until we have no more params
            let param;
            while (true) { // eslint-disable-line no-constant-condition
                param = paramSplit.exec(hash);
                if (!param) { break; }

                let key = param[1];
                let value = param[2];

                // id_token should remain base64url encoded
                if (key === 'id_token' || key === 'access_token' || key === 'code') {
                    obj[key] = value;
                } else {
                    obj[key] = decodeURIComponent(value.replace(plus2space, ' '));
                }
            }
            return obj;
        },
        objToFormUrlencoded:function(obj) {
            // Converts obj to be proper string for Content-Type "application/x-www-form-urlencoded"
            let str = [];
            for (let p in obj)
                str.push(encodeURIComponent(p) + "=" + encodeURIComponent(obj[p]));
            return str.join("&");
        },
        getRedirectUri: function(idp_id) {
            let path = window.location.origin + '/login/callback';
            let qs = '?idp_id=' + idp_id;
            return path + qs;
        }
    };
}])

.controller('LoginController', [
    '$scope', '$location', '$httpParamSerializer', '$window', 'slashDB', 'utils',
    function($scope, $location, $httpParamSerializer, $window, slashdb, utils) {
        let ctrl = this;
        function init () {
            // idps to remove from page
            ctrl.disabled_jwt_idps = [];
            ctrl.disabled_saml_idps = [];

            ctrl.slashdb = slashdb;
            ctrl.auth_settings = {};
            ctrl.user = '';
            ctrl.password = '';

            resetLoginErrors();
            if (!ctrl.loginSSOError) {
                // Don't overwrite referrer if handling error
                $window.sessionStorage.setItem('ssoApp.login_referrer', document.referrer);
            }
            slashdb.subscribeSettingsChange($scope, updateController);
            slashdb.setupSettings()
                .then(function (response) {
                    updateController();
                });
        }

        function resetLoginErrors() {
            ctrl.loginFormError = '';
            ctrl.loginApiKeyError = '';
            ctrl.loginSSOError = $window.sessionStorage.getItem('ssoApp.login_error');
            $window.sessionStorage.removeItem('ssoApp.login_error');
        }

        function updateController(){
            ctrl.auth_settings = angular.copy(slashdb.settings.auth_settings);
            ctrl.auth_policies = ctrl.auth_settings['authentication_policies']
            let auth_policies = ctrl.auth_policies;
            try {
                ctrl.visibleForm = auth_policies.login_form.enabled && auth_policies.login_form.gui.visible;
            } catch (error) {
                ctrl.visibleForm = false;
            }
            try {
                ctrl.visibleApikey = auth_policies.apikey.enabled && auth_policies.apikey.gui.visible;
            } catch (error) {
                ctrl.visibleApikey = false
            }
            try {
                ctrl.jwt_idps = {}
                for (const [idp_id, idp_config] of Object.entries(auth_policies.jwt.identity_providers)) {
                    if (idp_config.enabled && idp_config.gui.visible) {
                        ctrl.jwt_idps[idp_id] = idp_config
                    }
                }

                for (const unwanted_idp_id of ctrl.disabled_jwt_idps) {
                    delete ctrl.jwt_idps[unwanted_idp_id];
                }

                ctrl.visibleJwt = auth_policies.jwt.enabled && auth_policies.jwt.gui.visible &&
                    Object.keys(ctrl.jwt_idps).length > 0;
            } catch (error) {
                ctrl.visibleJwt = false;
                ctrl.jwt_idps = null
            }

            try {
                ctrl.saml_idps = {}
                for (const [idp_id, idp_config] of Object.entries(auth_policies.saml.identity_providers)) {
                    if (idp_config.enabled && idp_config.gui.visible) {
                        ctrl.saml_idps[idp_id] = idp_config
                    }
                }

                for (const unwanted_idp_id of ctrl.disabled_saml_idps) {
                    delete ctrl.saml_idps[unwanted_idp_id];
                }

                ctrl.visibleSaml = auth_policies.saml.enabled && auth_policies.saml.gui.visible &&
                    Object.keys(ctrl.saml_idps).length > 0;
            } catch (error) {
                ctrl.visibleSaml = false;
                ctrl.saml_idps = null;
            }
        }

        ctrl.login_with_jwt_idp = function(idp_id) {
            resetLoginErrors();
            $location.path('/login/sso/' + idp_id);
        };

        ctrl.login_with_saml_idp = function(idp_id) {
            resetLoginErrors();

            let ref = $window.sessionStorage.getItem('ssoApp.login_referrer');
            ref = (ref) ? ref : '/';

            window.location.href = `/sso/saml/${idp_id}?return_to=${ref}`;
        };

        ctrl.login_with_form = function () {
            resetLoginErrors();
            slashdb.login(ctrl.user, ctrl.password)
            .then(function (response) {
                utils.loginSuccessful();
            })
            .catch(function (response) {
                if (response.status === 429) {
                    ctrl.loginFormError = 'Too many unsuccessful logins. Please try again after several minutes.';
                } else {
                    ctrl.loginFormError = response.headers('Warning');
                }
            });
        };

        ctrl.login_with_apikey = function(){
            resetLoginErrors();
            let apikeyConf = ctrl.auth_settings.authentication_policies.apikey;
            let url = '/index.html?' + apikeyConf.apikey_name + "=" + ctrl.apikey;
            slashdb.get(url)
            .then(function(response){
                return slashdb.setupSettings()
            })
            .then(function(response){
                let curr_user = slashdb.settings.user;
                if (curr_user === null || curr_user === 'public'){
                    ctrl.loginApiKeyError = 'Login failed'
                } else {
                    utils.loginSuccessful();
                }
            })
        }

        init();
}])

.controller('SSOJwtController', [
    '$scope', '$location', '$httpParamSerializer', '$window', '$routeParams', 'slashDB', 'utils',
    function($scope, $location, $httpParamSerializer, $window, $routeParams, slashdb, utils) {
        let ctrl = this;

        function init() {
            ctrl.slashdb = slashdb;
            ctrl.idp_id = $routeParams.idp_id;
            ctrl.auth_settings = {};

            slashdb.subscribeSettingsChange($scope, updateController);

            slashdb.setupSettings()
                .then(function(response){
                    updateController();
                    login();
                })
        }

        function updateController(){
            ctrl.auth_settings = angular.copy(slashdb.settings.auth_settings);
            ctrl.auth_policies = ctrl.auth_settings['authentication_policies']
            let auth_policies = ctrl.auth_policies;
            try {
                ctrl.visibleForm = auth_policies.login_form.enabled && auth_policies.login_form.gui.visible;
            } catch (error) {
                ctrl.visibleForm = false;
            }
            try {
                ctrl.visibleApikey = auth_policies.apikey.enabled && auth_policies.apikey.gui.visible;
            } catch (error) {
                ctrl.visibleApikey = false
            }
            try {
                ctrl.visibleJwt = auth_policies.jwt.enabled && auth_policies.jwt.gui.visible;
                ctrl.jwt_idps = auth_policies.jwt.identity_providers;
            } catch (error) {
                ctrl.visibleJwt = false;
                ctrl.idp_id = null;
                ctrl.ssoConfig = null;
            }
        }

        function login() {
            let idp_id = ctrl.idp_id;
            let ssoConfig = ctrl.auth_policies.jwt.identity_providers[idp_id];
            // may happens if someone manually typed bad id
            if (ssoConfig === undefined) {
                $window.sessionStorage.setItem(
                    'ssoApp.login_error',
                    `Login failed, nonexistent IDP ${idp_id}`
                )
                utils.loginFailed()
            }
            let state = utils.generateRandomString(128);
            let nonce = utils.generateRandomString(128);
            let code_challenge_method = 'S256';
            let code_verifier = utils.generateRandomString(128);
            let code_challenge = utils.generateCodeChallenge(code_verifier);
            $window.sessionStorage.setItem('ssoApp.idp_id', idp_id);
            $window.sessionStorage.setItem('ssoApp.state', state);
            $window.sessionStorage.setItem('ssoApp.nonce', nonce);
            $window.sessionStorage.setItem('ssoApp.code_challenge_method', code_challenge_method);
            $window.sessionStorage.setItem('ssoApp.code_verifier', code_verifier);
            $window.sessionStorage.setItem('ssoApp.code_challenge', code_challenge);

            let qs = {
                'client_id': ssoConfig.client_id,
                'redirect_uri': utils.getRedirectUri(idp_id),
                'response_type': ssoConfig.response_type,
                'response_mode': 'fragment',
                'state': state,
                'nonce': nonce,
                'scope': ssoConfig.scope
            };

            if (ssoConfig.response_type === 'code') {
                qs['code_challenge'] = code_challenge;
                qs['code_challenge_method'] = code_challenge_method;
            }

            qs = $httpParamSerializer(qs);
            window.location = ssoConfig.authorization_endpoint + '?' + qs;
        }

        init();
}])

.controller('JwtCallbackController', [
    '$scope', '$location', '$window', '$http', 'slashDB', 'utils',
    function($scope, $location, $window, $http, slashdb, utils) {
        let ctrl = this;

        function init() {
            slashdb.subscribeSettingsChange($scope, updateController)

            slashdb.setupSettings()
                .then(function(response){
                    updateController();
                    authenticate();
                })
        }

        function updateController() {
            ctrl.auth_settings = angular.copy(slashdb.settings.auth_settings);
            ctrl.auth_policies = ctrl.auth_settings['authentication_policies'];
            ctrl.jwt_idps = ctrl.auth_policies.jwt.identity_providers
            ctrl.visibleForm = ctrl.auth_policies.login_form.enabled;
            ctrl.visibleApikey = ctrl.auth_policies.apikey.enabled;
            ctrl.visibleJwt = ctrl.auth_policies.jwt.enabled;
            ctrl.idp_id = $window.sessionStorage.getItem('ssoApp.idp_id');
            ctrl.ssoConfig = ctrl.jwt_idps[ctrl.idp_id];
        }

        function authenticate_with_id_token(id_token) {
            /* Use id_token to authenticate in SlashDB. */
            let auth_header = "Bearer " + btoa(id_token);
            return $http.get(
                '/index.html',
                {
                    'headers': {
                        'Authorization': auth_header,
                        'X-Identity-Provider-Id': ctrl.idp_id
                    }
                }
            )
            .then(function(response){
                return slashdb.setupSettings()
            })
            .then(function(response){
                let curr_user = slashdb.settings.user;
                if (curr_user === null || curr_user === 'public'){
                    $window.sessionStorage.setItem('ssoApp.login_error', 'Login failed');
                    utils.loginFailed();
                } else {
                    utils.loginSuccessful();
                }
            })
        }

        function authenticate_with_code(authorization_code) {
            /* Use authorization code from IDP to receive id_token from IDP
            to use that id_token to authenticate in SlashDB. */
            let payload = {
                'client_id': ctrl.ssoConfig.client_id,
                'code_verifier': $window.sessionStorage.getItem('ssoApp.code_verifier'),
                'redirect_uri': utils.getRedirectUri(ctrl.idp_id),
                'grant_type': 'authorization_code',
                'code': authorization_code
            }

            $http({
                method: 'POST',
                url: ctrl.ssoConfig.token_endpoint,
                headers: {'Content-Type': 'application/x-www-form-urlencoded'},
                transformRequest: utils.objToFormUrlencoded,
                data: payload
            }).then(function (response) {
                authenticate_with_id_token(response.data.id_token)
            });
        }

        function authenticate() {
            let auth_response = angular.extend(
                {},
                utils.hashToObject($location.hash()),
                $location.search()
            )

            if (auth_response.error !== undefined) {
                console.error("Error in authorization response: " + auth_response.error);
                console.error(auth_response.error_description);
                $location.path('/login/callback/failed');
                return null
            }

            let state = $window.sessionStorage.getItem('ssoApp.state');
            if (auth_response.state !== state) {
                console.error('Invalid PKCE state!');
                $location.search('error', 'invalid_pkce_state');
                $location.path('/login/callback/failed');
                return null;
            }

            if (auth_response.code !== undefined) {
                authenticate_with_code(auth_response.code);
            } else if (auth_response.id_token !== undefined) {
                authenticate_with_id_token(auth_response.id_token);
            } else {
                let error_description = "Missing authorization code or id_token in authorization response."
                console.error(error_description)
                $location.search('error', 'missing_auth_code');
                $location.search('error_description', error_description);
                $location.path('/login/callback/failed');
            }
        }

        init();
}])

.controller('SSOFailedController', [
    '$scope', '$location', '$window', '$http', 'slashDB', 'utils',
    function($scope, $location, $window, $http, slashdb, utils) {
        let ctrl = this;
        let auth_response = angular.extend(
            {},
            utils.hashToObject($location.hash()),
            $location.search()
        )
        $scope.idp_id = auth_response.idp_id;
        $scope.error = auth_response.error;
        $scope.error_description = auth_response.error_description;

}]);