'use strict';

angular.module('sdbgui')
    .component('databases', {
        templateUrl: '/static/components/sdbgui-databases/sdbgui-databases.template.html',
        controller: [
            '$q', '$scope', '$location', '$interval', '$window', '$uibModal', 'sortingService', 'slashDB', 'sessionControlService',
            function QueryController($q, $scope, $location, $interval, $window, $uibModal, sortingService, slashdb, sessionControlService) {
                const win = angular.element($window);
                const ctrl = this;
                ctrl.sort = sortingService;
                ctrl.connectStatusesInterval = null;
                ctrl.Statuses = {
                    Failed: 'Failed',
                    Connected: 'Connected',
                    Disconnected: 'Disconnected',
                    all: []
                };
                ctrl.Statuses.all = [ctrl.Statuses.Failed, ctrl.Statuses.Connected, ctrl.Statuses.Disconnected];

                ctrl.$onInit = function() {
                    $scope.searchable_fields = ['db_id', 'db_type', 'desc'];
                    ctrl.dbconfigs = [];
                    ctrl.settings=slashdb.settings;
                    ctrl.lockedActives = {};

                    win.on('focus', ctrl.onFocus);
                    win.on('blur', ctrl.onBlur);
                    $scope.updateConfigList();
                    updateConnectStatuses();

                    $scope.$on('$destroy', function () {
                        win.off('focus', ctrl.onFocus);
                        win.off('blur', ctrl.onBlur);
                        $interval.cancel(ctrl.connectStatusesInterval);
                    });
                }

                function getConnectStatuses(){
                    if (ctrl.dbconfigs.len === 0){
                        return null
                    }

                    slashdb.get('/settings/reflect-status.json', {}, false, false)
                        .then(function(response){
                            var connectStatuses = response.data;
                            angular.forEach(ctrl.dbconfigs, function (dbconf) {
                                dbconf.connect_status = connectStatuses[dbconf.db_id].model;
                                ctrl.lockedActives[dbconf.db_id] = false;
                            });
                        })
                        .catch(function(){
                            cancelConnectStatusInterval();
                        })
                }

                function updateConnectStatuses(){
                    if (ctrl.connectStatusesInterval == null){
                        ctrl.connectStatusesInterval = $interval(getConnectStatuses, 1000, 0);
                    }
                }

                function cancelConnectStatusInterval(){
                    $interval.cancel(ctrl.connectStatusesInterval);
                    ctrl.connectStatusesInterval = null;
                }

                ctrl.onFocus = function () {
                    getConnectStatuses();
                    // Starts an interval for connect status when the page is focused
                    updateConnectStatuses();
                }

                ctrl.onBlur = function () {
                    // Stops the interval for connect status when the page loses focus
                    cancelConnectStatusInterval();
                }

                $scope.updateConfigList = function () {
                    slashdb.get('/db.json?guidata', {}, false, false)
                        .then(function (response) {
                            let dbconfigs = [];
                            angular.forEach(response.data, function (new_dbconf, key) {
                                new_dbconf.db_id = key;
                                // keep old connect status to avoid flickering of status
                                const old_dbconf = ctrl.dbconfigs.find(function(element) {
                                    return element.db_id == new_dbconf.db_id
                                });
                                if (old_dbconf) {
                                    new_dbconf['connect_status'] = old_dbconf.connect_status;
                                }
                                dbconfigs.push(new_dbconf);
                            });
                            ctrl.listAlert = '';
                            ctrl.dbconfigs = dbconfigs;
                        })
                        .catch(function (response) {
                            ctrl.listAlert = response.headers('Warning');
                            ctrl.dbconfigs = [];
                        })
                        .finally(() => {
                            sessionControlService.onSessionLost(() =>
                                ctrl.listAlert = sessionControlService.expiredMsg
                            )
                            sessionControlService.onSessionRestore(()=>{
                                $scope.updateConfigList()
                            })
                        });
                };

                $scope.openMessageBox = function(messageBoxOptions){
                    const modalInstance = $uibModal.open({
                        templateUrl: '/static/components/sdbgui-dbconfigs/sdbgui-dbconfigs-messagebox.template.html',
                        size: 'lg',
                        windowTopClass: 'center-modal',
                        controller: 'DBConfigMessageBoxCtrl',
                        controllerAs: 'ctrl',
                        resolve: {
                            configId: function () {
                                return messageBoxOptions.configId;
                            },
                            showCachePrompt: function () {
                                return messageBoxOptions.showCachePrompt;
                            },
                            showInfo: function () {
                                return messageBoxOptions.showInfo;
                            },
                            hasShownInfo: function () {
                                return messageBoxOptions.hasShownInfo;
                            },
                            lockActive: function () {
                                // Lets us lock the power button without needing to send a request
                                return (db_id, value) => {
                                    ctrl.lockedActives[db_id] = value;
                                }
                            },
                        }
                    });
                    modalInstance.result
                        .finally(function (reason) {
                            // refresh config list after closing modal
                            $scope.updateConfigList();
                        });
                };

                $scope.toggleDBState = function(dbConfig) {
                    const currentConnectStatus = dbConfig.connect_status;
                    dbConfig.connect_status = '';
                    slashdb.get(`/db/${dbConfig.db_id}.json?guidata`, {}, false, false)
                        .then(function (response) {
                            const db_guidata = response.data;
                            const showCachePrompt = db_guidata.cacheable && db_guidata.cache_exists;
                            if (showCachePrompt && [ctrl.Statuses.Disconnected, ctrl.Statuses.Failed].includes(db_guidata.connect_status)) {
                                $scope.openMessageBox({
                                    configId: dbConfig.db_id,
                                    showCachePrompt: true,
                                    showInfo: false,
                                });
                            } else {
                                slashdb.toggleDB(dbConfig.db_id, currentConnectStatus)}
                        })
                        .finally(function () {
                            $scope.updateConfigList();
                        });
                };

                $scope.getStatusStyle = function (status) {
                    if (status === ctrl.Statuses.Connected) return 'bg-success';
                    if (status === ctrl.Statuses.Failed) return 'bg-danger';
                };

                ctrl.isLoading = function (dbConfig) {
                    return !ctrl.Statuses.all.includes(dbConfig.connect_status);
                };

                ctrl.isLocked = function (dbConfig) {
                    return ctrl.lockedActives[dbConfig.db_id];
                };
            }
        ]
    });
