( () => {
    angular.module('sdbgui')
        .provider('sessionControl', sessionLostInterceptor)
        .factory('handleExpiredSession', handleExpiredSession)
        .factory('sessionControlService', sessionControlService)
        .controller('LoginPopupBlackoutController', LoginPopupBlackoutController)

    function sessionLostInterceptor() {
        // no op, not intended to use as real provider
        this.$get = () => {}

        this.interceptor = ['$q', '$injector', 'handleExpiredSession',
            function ($q, $injector, handleExpiredSession) {
                return {
                    request: function (config) {
                        return config;
                    },
                    requestError: function (rejection) {
                        return $q.reject(rejection);
                    },
                    response: function (result) {
                        // exclude non json
                        if (!result.config.url.includes('.json')) return result

                        // exclude self
                        if (result.config.method === "GET"
                            && result.config.url.includes('/settings.json')) {
                            return result
                        }

                        // executed async, do not affect request duration
                        sessionControlService = $injector.get('sessionControlService')
                        sessionControlService.runOnlyIfSessionLost(() => handleExpiredSession())

                        return result;
                    },

                    responseError: function (response) {
                        // exclude self
                        if (response.config.method === "GET"
                            && response.config.url.includes('/settings.json')) {
                            return $q.reject(response);
                        }

                        // executed async, do not affect request duration
                        sessionControlService = $injector.get('sessionControlService')
                        sessionControlService.runOnlyIfSessionLost(() => handleExpiredSession())

                        return $q.reject(response);
                    }
                }
            }]
    }

    function centeredPopupWindow(url, title, w, h) {
        let y = window.outerHeight / 2 + window.screenY - (h / 2)
        let x = window.outerWidth / 2 + window.screenX - (w / 2)
        let params = 'toolbar=no, location=no, directories=no, status=no, ' +
            'menubar=no, scrollbars=no, resizable=no, copyhistory=no, ' +
            'width=' + w + ', height=' + h + ', top=' + y + ', left=' + x
        return window.open(url, title, params);
    }

    function getLoginPopup() {
        return centeredPopupWindow(
            '/login?close', 'login', 600, 750
        )
    }

    handleExpiredSession.$inject = ['$injector', '$window']
    function handleExpiredSession($injector, $window) {
        let overlayBlackoutOpened = false;

        return function () {
            const sendChannel = new BroadcastChannel('app-data')
            sendChannel.postMessage('sdb.sessionLost')
            
            if (overlayBlackoutOpened) {
                return
            }

            $window.localStorage.setItem('sdb.sessionLost', 'true');

            $uibModal = $injector.get('$uibModal')
            let overlayBlackout = $uibModal.open({
                windowClass: 'login-popup-blackout',
                templateUrl: '/static/components/sdbgui-login/login-popup.template.html',
                backdrop: 'static',
                controller: 'LoginPopupBlackoutController',
            })
            overlayBlackoutOpened = true;

            // close after login
            const recvChannel = new BroadcastChannel('app-data')
            let handler = (event) => {
                if (event.data !== 'sdb.loggedIn') return

                overlayBlackout.close()
                recvChannel.removeEventListener('message', handler)
            }
            recvChannel.addEventListener('message', handler)

            overlayBlackout.result.finally(() => {
                overlayBlackoutOpened = false;
            })
        }
    }


    LoginPopupBlackoutController.$inject = ['$uibModalInstance', '$scope', '$window']
    function LoginPopupBlackoutController($uibModalInstance, $scope, $window) {
        let $ctrl = this
        $scope.$ctrl = $ctrl

        $ctrl.onClose = () => {
            $uibModalInstance.dismiss()
        }

        $ctrl.cancel = () => {
            // delete markers
            $window.localStorage.removeItem('sdb.loggedIn');
            $window.localStorage.removeItem('sdb.sessionLost');

            // go to /
            window.location.assign("/");
        }

        $ctrl.loginPopup = () => {
            let loginPopup = getLoginPopup()
            loginPopup.focus()

            // close after login
            const recvChannel = new BroadcastChannel('app-data')
            let handler = (event) => {
                if (event.data !== 'sdb.loggedIn') return

                if (loginPopup !== null) {
                    loginPopup.close()
                }
                recvChannel.removeEventListener('message', handler)
            }
            recvChannel.addEventListener('message', handler)
        }
    }

    sessionControlService.$inject = ['$window', 'slashDB']
    function sessionControlService ($window, slashdb) {
        const recvChannel = new BroadcastChannel('app-data')

        let service = {}

        service.runOnlyIfSessionLost = (func) => {
            if ($window.localStorage.getItem('sdb.sessionLost') === 'true') {
                func()
            }
            else if ($window.localStorage.getItem('sdb.loggedIn') === 'true') {
                slashdb.setupSettings().then(() => {
                    if (['public', null].includes(slashdb.settings.user)) {
                        func()
                    }
                })
            }
        }

        service.onSessionLost = (func) => {
            service.runOnlyIfSessionLost(func)
        }

        service.onSessionRestore = (func) => {
            service.runOnlyIfSessionLost(() => {
                let handler = (event) => {
                    if (event.data !== 'sdb.loggedIn') return

                    func()
                    slashdb.setupSettings()
                    recvChannel.removeEventListener('message', handler)
                }
                recvChannel.addEventListener('message', handler)
            })
        }

        service.expiredMsg = 'Session expired'

        return service
    }
})();