class QuerystudioController
  @$inject: ['$window', '$location', '$timeout', '$rootScope', '$scope', 'slashDB',
    'QueryExecService', 'selectizeService', 'sessionControlService'
  ]
  constructor: (@$window, @$location, @$timeout, @$rootScope, @$scope, @slashdb,
    @QES, @selectizeService, @sessionControlService
  ) ->
    @tabs =
      side:
        config: true
        schema: false
      top:
        sql: true
        diagram: false
      bottom:
        execute: true
        results: false
    @required_fields = {}
    @executeOnly = null
    @codemirrorInstance = null
    @lastDbSchemaId = null
    @execUrls =
      html: ''
      json: ''
      csv: ''
      xml: ''
      xsd: ''
    # reinit forms on page restore from cache in FF
    @$window.onpageshow = (event) => if event.persisted then @$onInit()
    @fieldMappingQueryDef = {
      'query_id': 'Query ID',
      'database': 'Database',
      'read':'View',
      'write':'Edit',
      'execute':'Execute',
      'http_methods': 'HTTP Methods',
      'sqlstr': 'SQL Statement',
      'desc': 'Description',
    }
    @savingInProgress = false
    @executionInProgress = false
    @nullPlaceHolder = "<SDB_NULL>"
    @$window.addEventListener('beforeunload', (event) =>
      if @executionInProgress
        event.preventDefault()
        # Custom message is not supported on modern browsers.
    )

  $onInit: ->
    @slashdb.setupSettings()
    .then =>
      @initCodemirror()
      @initConfigForm()
      @initExecuteForm()
      @initResults()
      @initSchema()
      @updateConfigForm()
      .then =>
        @updateCodemirror()
        executeNow = @restoreExecuteForm()
        if executeNow
          @updateExecuteForm(@configData)
          res = @executeQuery()
          if not res
            @switchTab('bottom', 'execute')
    .finally =>
      # fix for renaming query when public in edit #1080
      if @slashdb.settings.user == 'public'
        @sessionControlService.onSessionLost =>
          setErrors(null, @sessionControlService.expiredMsg)

        @sessionControlService.onSessionRestore => @$onInit()

    return null

  initSchema: ->
    @$scope.$watch =>
      return @configData.database
    , (newValue, oldValue) =>
      if newValue and newValue != oldValue
        @updateHints(newValue)

  updateHints: (dbId) ->
    if not dbId
      return
    url = "/db/#{dbId}.json?tables"
    @slashdb.get url
    .then (response) =>
      schemaTables = response.data
      schemaHints = new Set([]);

      for table in schemaTables
        schemaHints.add table.name
        for column in table.columns
          schemaHints.add column.name

      @schemaHints = Array.from(schemaHints)

  initConfigForm: ->
    @$scope.databases = []
    @slashdb.get('/db.json', { params: guidata: '' }, false, false).then (response) =>
      for dbId, dbInfo of response.data
        dbInfo.db_id = dbId
        @$scope.databases.push dbInfo
      return
    # stores database definition model, some initial values
    @configData = editable: true

    @updateSuggestedMethods()
    @$scope.newConfig = !Boolean(@queryId)
    @$scope.configId = @queryId
    @$scope.okVisible = false
    @$scope.errorMessage = ''
    @$scope.databases = []


  initExecuteForm: ->
    @httpMethods = ['GET', 'POST', 'PUT', 'DELETE']
    @extensions = ['html', 'json', 'csv', 'xml', 'xsd']

    @defaultLimit = parseInt(@slashdb.settings.html_limit) or 29

    @ready = false
    @limit = @defaultLimit
    @offset = 0
    @count = false
    @format = 'json'
    @distinct = false
    @transpose = false
    @stream = false
    @nullStr = false
    @sortMap = new Map
    @execUrl = ''
    @execPath = ''
    @queryExecuted = false
    @selectedHTTPMethod = null
    @queryParameters = {}
    @parameterNullState = {}

    @$scope.methodAvailable = (m) =>
      http_methods = @configData.http_methods
      if http_methods
        return http_methods[m]
    @$scope.hasParameters = =>
      return Object.keys(@queryParameters).length > 0

  initResults: ->
    @queryOutput = null

    @$rootScope.$on 'queryExecuted', (event, args) =>
      @executionInProgress = false
      @queryOutput = @QES.queryOutput
      httpResponse = args.httpResponse
      @resultsResponseCode = httpResponse.status
      @resultsResponseMessage = "#{httpResponse.status} #{httpResponse.statusText}"
      @resultsResponseWarning = httpResponse.headers 'Warning'
      @resultsCount = httpResponse.headers(
        "#{@slashdb.settings.branding_name}-all-record-count"
      )
      @resultsCountFailed = (@resultsResponseWarning == 'Failed to get count.')

      noData = (httpResponse.data? and httpResponse.data.length == 0) or not httpResponse.data?
      if httpResponse.status < 400 and noData
        @resultsResponseNoMoreData = true
      else
        @resultsResponseNoMoreData = false

      if httpResponse.status < 400
        @transposeResultTable = @transpose
        if @transpose
          @queryOutput.data = @queryOutput.data[0]

      @switchTab 'bottom', 'results'

  initCodemirror: ->
    handleLoadedEditor = (_editor) =>
      @codemirrorInstance = _editor
      _editor.on 'change', =>
        @updateSuggestedMethods()

    produceHints = (mc, opts) =>
      hints = CodeMirror.hint.sql(mc, opts)
      dbschemaHints = CodeMirror.hint.dbschema(mc, opts, @schemaHints)

      if dbschemaHints
        for hint in dbschemaHints.list
          hints.list.push hint

      return hints

    # Has to change in value to apply changes in CodeMirror settings
    @$scope.refreshCodemirror = true
    @$scope.editorOptions =
      onLoad: handleLoadedEditor
      lineNumbers: true
      readOnly: false
      mode: 'text/x-sql'
      hintOptions:
        hint: produceHints
      extraKeys:
        "Ctrl-Space": "autocomplete"

  updateCodemirror: ->
    currDbType = @$scope.databases?[@configData.database]?.db_type
    cmMode = switch currDbType
      when 'sqlite' then "text/x-sqlite"
      when 'mssql' then 'text/x-mssql'
      when 'oracle' then 'text/x-plsql'
      when 'mysql' then 'text/x-mysql'
      when 'postgresql' then 'text/x-gpsql'
      when 'db2' then 'text/x-sql'
      else 'text/x-sql'

    @$scope.editorOptions.mode = cmMode
    @$scope.editorOptions.readOnly = if !@configData?.editable then 'nocursor' else false
    @$scope.refreshCodemirror = !@$scope.refreshCodemirror

  switchTab: (section, tabName) ->
    for tab in (Object.keys @tabs[section])
      @tabs[section][tab] = tab == tabName

  updateConfigForm: ->
    if @$scope.newConfig
      @executeOnly = false
      @configData.executable = true
      @updateCodemirror()
      @selectizeService.updateUserOptions null
    else
      @$scope.errorMessage = ''
      promise = @slashdb.get '/querydef/' + @$scope.configId + '.json', { params: guidata: '' }, false, false
      .then (response) =>
        @configData = angular.copy(response.data)
        @executeOnly = @configData.executable and !(@configData.editable or @configData.viewable)
        if @configData.editable or @configData.viewable
          @selectizeService.updateUserOptions @configData

        @updateSuggestedMethods()
        @updateCodemirror()
        @updateExecuteForm(@configData)
        return
      .catch (response) =>
          @$scope.errorMessage = response.headers('Warning')

      return promise

  clean: (andDisable) ->
    @configData = editable: !andDisable
    return

  delete: ->
    @$scope.errorMessage = ''
    @slashdb.deleteQueryDef(@$scope.configId)
    .then (response) =>
      @slashdb.setupSettings()
      @$window.location.href = '/querydef'
      return
    .catch (response) =>
      @$scope.errorMessage = response.headers('Warning')
      return
    .finally =>
      @sessionControlService.onSessionLost =>
        setErrors(null, @sessionControlService.expiredMsg)

      @sessionControlService.onSessionRestore(() =>
        # update current user after login
        @slashdb.setupSettings()
        @resetErrors()
      )
    return

  saveExecuteForm: (execute)->
    formParams =
      format: @format
      limit: @limit
      offset: @offset
      count: @count
      distinct: @distinct
      transpose: @transpose
      stream: @stream
      selectedHTTPMethod: @selectedHTTPMethod
      queryParameters: @queryParameters
      queryId: @configData.query_id
      executeNow: execute

    try
      localStorage.setItem('qsExecuteQuery', JSON.stringify(formParams))
    catch e
      if e == QUOTA_EXCEEDED_ERR
        alert('local storage limit exceeded')

  restoreExecuteForm: ->
    formParams = JSON.parse(localStorage.getItem('qsExecuteQuery'))
    if not formParams?
      return false
    localStorage.removeItem('qsExecuteQuery')

    if formParams.queryId != @queryId
      return false

    @limit = formParams.limit
    @offset = formParams.offset
    @count = formParams.count
    @distinct = formParams.distinct
    @transpose = formParams.transpose
    @stream = formParams.stream
    @format = formParams.format
    @selectedHTTPMethod = formParams.selectedHTTPMethod
    @queryParameters = formParams.queryParameters

    return formParams.executeNow

  setRequiredFields: (errors) ->
    required = {}
    for err in errors
      if err.type == 'MISSING'
        required[err.field] = true
      # treat empty sqlstr as required
      if err.field == 'sqlstr' and err.type == 'INVALID'
        required[err.field] = true
      # treat empty query id as required
      if err.field == 'query_id' and err.type == 'INVALID' and not @configData.query_id
        required[err.field] = true
      # http methods can only be invalid when missing
      if err.field == 'http_methods'
        required[err.field] = true

    ['query_id', 'database', 'http_methods', 'sqlstr'].forEach((value, _, __) ->
      required[value] = true
    )
    @required_fields = required

  setInvalidFields: (errors) ->
    for err in errors
      if err.type == 'INVALID' and err.field of @configForm
        # treat empty query id as required, but not invalid
        if err.field == 'query_id' and not @configData.query_id
          continue
        # treat empty sqlstr as required, but not invalid
        if err.field == 'sqlstr' and err['description'].includes('is too short')
          continue

        @configForm[err.field].$setValidity('valid', false)

  getUserFriendlyErrorMessages: (errors) ->
    return 'Errors found in incoming query definition: ' + errors.map(
      (err) =>
        if err['type'] == 'MISSING'
          if err['field'] == 'http_methods'
            return "At least one HTTP Method must be selected"
          return "#{ @fieldMappingQueryDef[err['field']] } is obligatory"

        if err['type'] == 'EXTRA'
          return "#{ err['description'] } - try to reload page"

        if err['field'] == 'sqlstr' and err['description'].includes('is too short')
          return "#{ @fieldMappingQueryDef[err['field']] } is required"

        if err['field'] == 'query_id' and not @configData.query_id
          return "#{ @fieldMappingQueryDef[err['field']] } is required"

        if err['field'] == 'query_id' and err['description'].includes('already exists')
          return err['description']


        if not @configData.http_methods? or Object.keys(@configData.http_methods).length == 0
          isEmptyHttpMethods = true
        else
          isEmptyHttpMethods = false

        if err['field'] == 'http_methods' and isEmptyHttpMethods
          return "At least one HTTP Method must be selected"

        if err['field'] in ['execute', 'read', 'write'] and err['description'].includes('is not valid under')
          return "'#{ @fieldMappingQueryDef[err['field']] }': #{ err['description'] }"

        if err['description'].includes('is too long')
          return "#{ @fieldMappingQueryDef[err['field']] } is too long"
        return "#{ @fieldMappingQueryDef[err['field']] } is invalid"
    ).join('; ') + '.'


  resetErrors: () ->
    @$scope.errorMessage = ''
    @required_fields = {}

    for key, val of @configForm
      if typeof val == 'object' and val.hasOwnProperty('$modelValue')
        val.$setValidity('valid', true)

  resetError: (field) ->
      @configForm[field].$setValidity('valid', true)

  setErrors: (errors, description) ->
    if errors
      @$scope.errorMessage = @getUserFriendlyErrorMessages(errors)
      @setInvalidFields(errors)
      @setRequiredFields(errors)
    else
      @$scope.errorMessage = description


  getAvaliableHttpMethods: (http_methods) ->
    available_methods = {}
    for own m, b of http_methods
      if b
        available_methods[m] = true

    return available_methods

  getPayload: (raw_data) ->
    payload = Object.assign({}, raw_data)
    delete payload.editable
    delete payload.executable
    delete payload.viewable
    delete payload.url_template
    delete payload.create
    return payload

  save: (execute = false) ->
    @savingInProgress = true
    if not @configData.read then @configData.read = []
    if not @configData.write then @configData.write = []
    if not @configData.execute then @configData.execute = []

    @resetErrors()

    payload = @getPayload(@configData)

    if @$scope.newConfig
      promise = @slashdb.createQueryDef(payload)
      promise.then (response) =>
        @saveExecuteForm(execute)
        @goToNewEndpoint()
        @setRequiredFields([])
        return
      .catch (response) =>
        @setErrors(response.data.errors, response.data.description)
        return
    else
      dirtyUrl = @configForm.query_id.$dirty
      promise = @slashdb.updateQueryDef(@$scope.configId, payload)
      promise.then (response) =>
        @blinkOK()
        @$scope.configId = @configData.query_id
        @updateConfigForm()
        if dirtyUrl
          @saveExecuteForm(execute)
          @goToNewEndpoint()
        @setRequiredFields([])
        return
      .catch (response) =>
        @setErrors(response.data.errors, response.data.description)
        return

    promise
      .finally =>
        @savingInProgress = false
        return
      .finally =>
        @sessionControlService.onSessionLost =>
          @setErrors(null, @sessionControlService.expiredMsg)

        @sessionControlService.onSessionRestore =>
          # update user
          @slashdb.setupSettings();
          @resetErrors()

    return promise

  goToNewEndpoint: ->
    @$window.location.assign("/querystudio/#{@configData.query_id}")

  saveAndExecute: ->
    @save(execute=true)
    .then (sdbResponse) =>
      if not @$scope.newConfig and sdbResponse.status >= 200 and sdbResponse.status < 400
        return @updateConfigForm()
          .then =>
            res = @executeQuery()
            if not res
              @switchTab('bottom', 'execute')

  isPrevPageEnabled: ->
    return not (@resultsResponseCode >= 400 or @offset <= 0 or @executionInProgress or @savingInProgress)

  previousResultPage: ->
    @offset -= @limit
    if @offset < 0 then @offset = 0
    @executeQuery()
    return

  isNextPageEnabled: ->
    return not(@resultsResponseCode >= 400 or !@queryOutput or
      @getQueryOutputSize() < @limit or @executionInProgress or @savingInProgress)

  nextResultPage: ->
    @offset += @limit
    @executeQuery()
    return

  updateExecUrl: ->
    if @configData

      @checkParamNullState()
      @validate()
      if @ready
        for key of @execUrls
          urlPath = @QES.getExecPath(@configData.query_id, @queryParameters,
            @limit, @offset, @count, key, @distinct, @transpose, @stream, @nullStr, @mapToSortArray(@sortMap))
          @execUrls[key] = @QES.queryBasePath + urlPath
      else
        for key of @execUrls
          @execUrls[key] = ''

      url = @QES.getExecPath(@configData.query_id, @queryParameters, @limit,
        @offset, @count, @format, @distinct, @transpose, @stream, @nullStr, @mapToSortArray(@sortMap))
      @execUrl = @QES.queryBasePath + url
      @execPath = url
      return @execUrl
    return

  setParamNull: (paramKey) ->
    if @parameterNullState[paramKey] == true
      @parameterNullState[paramKey] = false
      @queryParameters[paramKey] = ''  
    else
      @parameterNullState[paramKey] = true
      @queryParameters[paramKey] = @slashdb.settings['sqlpassthru.null_str']

  checkParamNullState: () ->
    # check if any of the parameters are currently set to the null placeholder typed in as a string
    # if so, use nullStr query parameter string to define a custom placeholder
    for key of @queryParameters
      if @queryParameters[key] == @slashdb.settings['sqlpassthru.null_str'] and @parameterNullState[key] != true
        @nullStr = true
        break
      @nullStr = false

    # when null placeholder value is entered as a string:
    # pass over existing nulls and set them to the custom placeholder
    if @nullStr == true
      for key of @queryParameters
        if @parameterNullState[key] == true
          @queryParameters[key] = @nullPlaceHolder

    # if an existing string value containing the null placeholder is changed to something else:
    # pass over existing nulls and set them to default placeholder
    else if @nullStr == false
      for key of @queryParameters
        if @parameterNullState[key] == true
          @queryParameters[key] = @slashdb.settings['sqlpassthru.null_str']

  executeQuery: ->
    @validate()
    if @ready
      queryStringParams =
        limit: @limit
        offset: @offset
        columns: true
        numeric_columns: true

      if @distinct
        queryStringParams.distinct = ''

      if @transpose
        queryStringParams.transpose = ''

      if @stream
        queryStringParams.stream = true

      if @sortMap.size
        queryStringParams.sort = @mapToSortArray(@sortMap)

      if @count
        queryStringParams['count']=''

      if @nullStr
        queryStringParams['nullStr'] = @nullPlaceHolder

      if Object.keys(@queryParameters).length == 0
        queryParams = null
      else if @selectedHTTPMethod in ['PUT', 'POST']
        queryParams = [@queryParameters]
      else
        queryParams = @queryParameters
      @executionInProgress = true
      @QES.executeQuery @configData.query_id, @selectedHTTPMethod, queryParams, queryStringParams
      .finally =>
        @sessionControlService.onSessionLost =>
          @resultsResponseWarning = @sessionControlService.expiredMsg

        @sessionControlService.onSessionRestore =>
          @queryOutput = null
          @resultsResponseCode = undefined
          @resultsResponseMessage = undefined
          @resultsResponseWarning = undefined
          @resultsCount = undefined
          @resultsCountFailed = undefined


    return @ready

  validate: ->
    valid = true
    valid = valid and @limit > -1
    valid = valid and @format in @extensions
    valid = valid and @selectedHTTPMethod in @httpMethods

    if @executeForm
      for key of @queryParameters
        formField = @executeForm[key]
        if formField
          formField.$touched = true
          valid = valid and formField.$valid
    # page not loaded yet, just check that params are present
    else
      for key, value of @queryParameters
        valid = valid and value

    @ready = valid

  blinkOK: ->
    @$scope.okVisible = true
    @$timeout =>
      @$scope.okVisible = false
      return
    , 1000
    return

  methodsMapping =
    'select':
      re: /\bselect\b/i
      method: 'GET'
    'insertOrExec':
      re: /\binsert|exec\b/i
      method: 'POST'
    'update':
      re: /\bupdate\b/i
      method: 'PUT'
    'delete':
      re: /\bdelete\b/i
      method: 'DELETE'
  methodsMappingKeys = Object.keys(methodsMapping)
  methodsMappingKeysL = methodsMappingKeys.length

  updateSuggestedMethods: ->
    suggestedMethods =
      'GET': false
      'POST': false
      'PUT': false
      'DELETE': false
    if @configData == null
      return suggestedMethods
    sqlstr = @configData.sqlstr
    searchTerm = undefined
    i = 0
    while i < methodsMappingKeysL
      searchTerm = methodsMapping[methodsMappingKeys[i]]
      suggestedMethods[searchTerm.method] = searchTerm.re.exec(sqlstr) != null
      i++
    @$scope.suggestedMethods = suggestedMethods

  getSqlstrParams: (sqlstr) ->
    allParams = []
    allParamsRegex = /:([\w\$]+)(?![:\w\$])/g
    match = allParamsRegex.exec sqlstr
    while match
      paramName = match[1]
      allParams.push paramName
      match = allParamsRegex.exec sqlstr

    excludedParams = []
    excludedParamsRegex = /[:\w\$\x5c]:([\w\$]+)(?![:\w\$])/g
    match = excludedParamsRegex.exec sqlstr
    while match
      paramName = match[1]
      excludedParams.push paramName
      match = excludedParamsRegex.exec sqlstr

    return $(allParams).not(excludedParams).get()

  updateExecuteForm: (query) ->
    oldParameters = angular.copy @queryParameters
    availableMethods = (m for m, v of query.http_methods when v is true)
    @updateSuggestedMethods()
    @queryParameters = {}

    # only when user have view access
    # params from editor
    if query.sqlstr
      for paramName in @getSqlstrParams query.sqlstr
        if oldParameters.hasOwnProperty paramName
          paramValue = oldParameters[paramName]
        else
          paramValue = ''
        @queryParameters[paramName] = paramValue

    # when user do not have view access
    # params from API
    if query.parameters
      for paramName in query.parameters
        if oldParameters.hasOwnProperty paramName
          paramValue = oldParameters[paramName]
        else
          paramValue = ''

        @queryParameters[paramName] = paramValue

    # preselect some http method for the form. First select by suggested then by available.
    if @selectedHTTPMethod not in availableMethods
      for m in @httpMethods
        if m in availableMethods
          if @$scope.suggestedMethods[m]
            @selectedHTTPMethod = m
            break

    if @selectedHTTPMethod not in availableMethods
      for m in @httpMethods
        if m in availableMethods
          @selectedHTTPMethod = m
          break

    return

  newQuery: ->
    if @configForm.$dirty or @sqlDirty
      @confirmNewQueryVisible = true
    else
      @openNewQueryPage()

  openNewQueryPage: ->
      window.location.assign("/querystudio/")

  sqlKeyDown: (e) ->
    if @configData.executable and (e.ctrlKey or e.metaKey) and (e.keyCode == 13 or e.keyCode == 10)
      @saveAndExecute()

  sqlChange: (e) ->
    if @configData.executable
      @updateExecuteForm(@configData)
      @sqlDirty = true

  onHttpMethodChange: ->
    if @configData.executable
      @configData.http_methods = @getAvaliableHttpMethods(
        @configData.http_methods
      )
      @updateExecuteForm(@configData)

  onSort: (column) ->
    if @executionInProgress
        return
    if column == null
       @sortMap.clear()
    else if not @sortMap.has(column)
      @sortMap.set(column, true)
    else
      v = @sortMap.get(column)
      if v
        @sortMap.set(column, false)
      else
        @sortMap.set(column, true)


    @executeQuery()

  mapToSortArray: (sortMap) ->
    return Array.from(sortMap, ([k, v]) ->
        if v
          return k
        else
          return "-#{k}"
    ).join(',')

  getSortCssClass: (column) ->
    if @sortMap.has(column)
      v = @sortMap.get(column)
      if v
        return '-up'
      else
        return '-down'
    else
      return ''

  getQueryOutputSize: () ->
    if @transpose
      col = @queryOutput.columnNames[0]
      try
        return @queryOutput.data[col].length
      catch error
        return 0

    else
      return @queryOutput.data.length

  
  getResultElementClass: (value, col) ->
    classes = ""
    if @queryOutput.numericColumns[col]
      classes += "numeric-val "
    if value == null
      classes += "null-val "
    if value == ""
      classes += "empty-val "

    return classes.trim()

  getResultValue: (value) ->
    decimalDigits = @slashdb.settings['html_sdb.float_decimal_digits']

    if value == null
       return @slashdb.settings['html_sdb.sqlpassthru.null_display_str']
    
    if value == ""
      return @slashdb.settings['html_sdb.sqlpassthru.empty_display_str']

    # check if integer or float
    if (
      Number(decimalDigits) == decimalDigits and decimalDigits % 1 == 0 and 
      Number(value) == value and value % 1 != 0
    )
      return Math.round(value*Math.pow(10,decimalDigits))/Math.pow(10,decimalDigits);
    else
      return value

angular.module 'sdbgui'
.component 'querystudio', {
  templateUrl: '/static/components/sdbgui-querystudio/querystudio.html'
  controller: QuerystudioController
  bindings:
    queryId: '@'
}
